-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getnotehistory.sql, tbsm_dev_tedw, tbsm_base31e
-- SID: 1.18
-- Date Modified: 07/03/19 15:53:56

-- Module description:
-- Create the etl_getnotehistory procedure 

USE Object
IF EXISTS (SELECT * FROM sysobjects WHERE type = 'P' AND name = 'etl_getnotehistory')
  DROP PROCEDURE etl_getnotehistory
GO

-- File: archive/db/sql/procs/etl_getnotehistory.sql, tbsm_dev_tedw, tbsm_base31e
-- SID: 1.18
-- Date Modified: 07/03/19 15:53:56

-- Module description:
-- Create the etl_getnotehistory procedure which is used by
-- TDW to receive a result set of note history records.
-- The procedure also adds to the twh_defer_notehistory table for note history data
-- that does not have a resource in twh_Inventory.
-- The procedure also updates the notehistory row of twh_ExtractCtrl: mtime is set to the current time
-- and status is changed to 1

CREATE PROCEDURE etl_getnotehistory
  @verbose TINYINT=0
AS
  SET NOCOUNT ON

  DECLARE @now NVARCHAR(40)
  DECLARE @rc INT
  DECLARE @err INT
  DECLARE @sql NVARCHAR(2000)
  DECLARE @this_xid DECIMAL(18)
  DECLARE @last_xid DECIMAL(18)
  DECLARE @gmoffsetmin INT
  DECLARE @last_gmoffsetmin INT
  DECLARE @GMT_CHANGE_TIME DATETIME
  DECLARE @process_gmtc TINYINT

  CREATE TABLE #notehistory_set (
    notehis_id       INT NOT NULL,            -- ident column from NoteHistory
    noteid           INT NOT NULL,            -- Note identifier From NoteHistory
    Action           NVARCHAR(255) NOT NULL,  -- From NoteHistory
    Executer         NVARCHAR(255) NULL,      -- From NoteHistory
    TimeExecuted     DATETIME NOT NULL,       -- From NoteHistory
    event_type       NVARCHAR(4) NOT NULL,    -- `Note'
    name             ObjName NOT NULL,        -- Name associated with note From Note_V
    Subject          NVARCHAR(255) NULL,      -- Subject of note From NoteHistoryDetail
    Body             NVARCHAR(1340) NULL,     -- Body of note From NoteHistoryDetail
    note_type        NVARCHAR(254) NOT NULL,  -- Type of note _Type from Note_V
    sevt_corr_val    NVARCHAR(254) NULL,      -- If there is a row from NoteDependent, the value is nd.eventcid + `\' 
                                              -- If there is no row from NoteDependent, the value is NULL
    twh_sevt_id      INT,                     -- 0
    comp_id          DECIMAL(20,0) NOT NULL,  -- This will come from the comp_id column of the twh_Inventory table, based 
                                              -- on the CreatorCID and CreatorID from the Note_V table.
                                              -- The value can be zero, which means that the resource does not yet appear
                                              -- in twh_Inventory and needs to be deferred.  A row can also have a non-zero
                                              -- value, which means it was included in the previous result set but hasn't been
                                              -- removed from the deferred table yet because TDW hasn't told us the that result
                                              -- set was successfully received and processed.
    twh_tevt_id      INT NOT NULL,            -- 0
    evt_corr_val     NVARCHAR(40) NOT NULL    -- `NOTE\' + CONVERT(NVARCHAR(40), nh.ident)
  )

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory starts', 0, 1, @now) WITH NOWAIT
  END

  IF @verbose >= 2 
  BEGIN
    SELECT 'twh_ExtractCtrl'='twh_ExtractCtrl',* FROM twh_ExtractCtrl WHERE name = 'notehistory'
  END

  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'notehistory' AND status=0) 
  BEGIN
    -- if status=0, TDW told TBSM that the previous result set was successfully processed so we can
    -- delete any deferred rows we sent last time
    DELETE twh_defer_notehist WHERE comp_id <> 0
    SELECT @rc = @@ROWCOUNT, @err = @@ERROR

    IF @err < 0 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getnotehistory: error %d deleting sent rows from twh_defer_notehist', 11, 1, @now, @err) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getnotehistory deleted %d rows from twh_defer_notehist that were sent to TDW last time', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  ELSE 
  BEGIN
    -- if status<>0, TDW did NOT tell TBSM that the previous result set was successfully processed.  We
    -- need to try again but we don't want to defer those notehistory that we deferred last time.  We'll start
    -- out by forgetting those we tried to send last time.
    DELETE twh_defer_notehist WHERE age <= 1
    SELECT @rc = @@ROWCOUNT, @err = @@ERROR

    IF @err < 0 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getnotehistory: error %d deleting rows last deferred from twh_defer_notehist', 11, 1, @now, @err) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getnotehistory deleted %d rows from twh_defer_notehist that were deferred from last time', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- indicate that a result set has been prepared for TDW.
  UPDATE twh_ExtractCtrl SET status = 1 WHERE name = 'notehistory'
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory: error %d updating twh_ExtractCtrl', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  -- add rows to twh_deffered_notehistory first until we know their TDW comp_id
  -- handle with NoteDependent
  SET @sql = ' ' +
  'INSERT INTO twh_defer_notehist ' +
  '  SELECT nh.ident AS notehis_id, ' +
  '         nh.noteid AS noteid, ' +
  '         CASE WHEN nh.Action IS NULL OR nh.Action = '''' THEN ''no action'' ELSE nh.Action END AS action,  ' +
  '         CASE WHEN nh.Executer IS NULL OR nh.Executer = '''' THEN ''no executer'' ELSE nh.Executer END AS executer,  ' +
  '         nh.TimeExecuted AS TimeExecuted, ' +
  '         ''Note'' AS event_type, ' +
  '         CASE WHEN n.name IS NULL OR n.name = '''' THEN ''no name'' ELSE n.name END AS name, ' +
  '         CASE WHEN nhd.Subject IS NULL OR nhd.Subject = '''' THEN ''no subject'' ELSE nhd.Subject END AS Subject, ' +
  '         CASE WHEN nhd.Body IS NULL OR nhd.Body LIKE '''' THEN ''no body'' ELSE substring(nhd.Body,0,1300) END AS body, ' +
  '         n._Type AS note_type, ' +
  '         nd.eventcid + ''\'' + CAST(nd.eventid AS NVARCHAR(40)) AS sevt_corr_val, ' +
  '         0 AS sevt_corr_id, ' +
  '         0, ' +
  '         0 AS twh_tevt_id, ' +
  '         ''NOTE\'' + CAST(nh.ident AS NVARCHAR(40)), ' +
  '         CASE WHEN nd.depcid IS NOT NULL THEN nd.depcid ELSE n._CreatorCID END, ' +
  '         CASE WHEN nd.depid  IS NOT NULL THEN nd.depid  ELSE n._CreatorObjID END, ' +
  '         0 ' +
  '    FROM NoteHistory              nh  WITH (NOLOCK) ' +
  '    JOIN Note_V                   n   WITH (NOLOCK) ON n.id = nh.noteid ' +
  '    JOIN NoteHistoryDetail        nhd WITH (NOLOCK) ON nhd.NoteHistory_id = nh.ident ' +
  '    JOIN NoteDependent            nd  WITH (NOLOCK) ON nd.noteid = nh.noteid'

  SELECT @this_xid = this_xid, @last_xid = last_xid FROM twh_ExtractCtrl WHERE name = 'notehistory'
  SET @sql = @sql + 
   ' AND nh.ident >  ' + CAST(@last_xid AS NVARCHAR(18)) +
   ' AND nh.ident <= ' + CAST(@this_xid AS NVARCHAR(18))

  IF @verbose >= 2 
  BEGIN
    PRINT @sql
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory executing: %s', 0, 1, @now, @sql) WITH NOWAIT
  END
  EXEC(@sql)
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory: error %d inserting into twh_defer_notehist with NoteDependent', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory inserted %d rows into twh_defer_notehist', 0, 1, @now, @rc) WITH NOWAIT
  END

  --
  -- handle without NoteDependent
  SET @sql = 
  ' INSERT INTO twh_defer_notehist ' +
  '   SELECT nh.ident AS notehis_id, ' +
  '      nh.noteid AS noteid, ' +
  '      CASE WHEN nh.Action IS NULL OR nh.Action = '''' THEN ''no action'' ELSE nh.Action END AS action,  ' +
  '      CASE WHEN nh.Executer IS NULL OR nh.Executer = '''' THEN ''no executer'' ELSE nh.Executer END AS executer,  ' +
  '      nh.TimeExecuted AS TimeExecuted, ' +
  '      ''Note'' AS event_type, ' +
  '      CASE WHEN n.name IS NULL OR n.name = '''' THEN ''no name'' ELSE n.name END AS name, ' +
  '      CASE WHEN nhd.Subject IS NULL OR nhd.Subject = '''' THEN ''no subject'' ELSE nhd.Subject END AS Subject, ' +
  '      CASE WHEN nhd.Body IS NULL OR nhd.Body LIKE '''' THEN ''no body'' ELSE substring(nhd.Body,0,1300) END AS body, ' +
  '      n._Type AS note_type, ' +
  '      '''' AS sevt_corr_val, ' +
  '      0 AS sevt_corr_id, ' +
  '      0, ' +
  '      0 AS twh_tevt_id, ' +
  '      ''NOTE\'' + CAST(nh.ident AS NVARCHAR(40)), ' +
  '      n._CreatorCID, ' +
  '      n._CreatorObjID, ' +
  '      0 ' +
  '   FROM NoteHistory              nh  WITH (NOLOCK) ' +
  '   JOIN Note_V                   n   WITH (NOLOCK) ON n.id = nh.noteid ' +
  '   JOIN NoteHistoryDetail        nhd WITH (NOLOCK) ON nhd.NoteHistory_id = nh.ident ' +
  '   WHERE noteid NOT IN (SELECT DISTINCT noteid FROM NoteDependent WITH (NOLOCK))'

  SELECT @this_xid = this_xid, @last_xid = last_xid FROM twh_ExtractCtrl WHERE name = 'notehistory'
  SET @sql = @sql + 
    ' AND nh.ident >  ' + CAST(@last_xid AS NVARCHAR(18)) + 
    ' AND nh.ident <= ' + CAST(@this_xid AS NVARCHAR(18))

  IF @verbose >= 2 
  BEGIN
    PRINT @sql
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory executing: %s', 0, 1, @now, @sql) WITH NOWAIT
  END
  EXEC(@sql)
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory: error %d inserting into twh_defer_notehist without NoteDependent', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory inserted %d rows into twh_defer_notehist', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- backfill TDW comp_id's - this may hit rows we just added or rows that were deferred from
  -- previous ETLs
  UPDATE h SET comp_id = i.comp_id
    FROM twh_defer_notehist h
    JOIN twh_Inventory      i ON i.cid = h.cid AND i.id = h.id
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory: error %d updating comp_id''s in twh_defer_notehist', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory updated comp_id of %d rows of twh_defer_notehist', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- adjust timestamps in result set to be GMT-based
  EXEC etl_getgmtvars 'notehistory', @process_gmtc OUTPUT, @gmoffsetmin OUTPUT, @last_gmoffsetmin OUTPUT, @GMT_CHANGE_TIME OUTPUT, @verbose=@verbose

  IF @process_gmtc = 0 -- adjust by a single offset
    UPDATE twh_defer_notehist SET TimeExecuted = DATEADD(MI, @gmoffsetmin, TimeExecuted) WHERE age = 0
  ELSE                 -- adjust by the two offsets
    UPDATE twh_defer_notehist SET TimeExecuted = CASE
                                                   WHEN TimeExecuted >= @GMT_CHANGE_TIME THEN DATEADD(MI, @gmoffsetmin, TimeExecuted)
                                                   ELSE                                       DATEADD(MI, @last_gmoffsetmin, TimeExecuted)
                                                 END
           WHERE age = 0

  -- Pull out the rows that have TDW comp_id's - these are what we'll return in the result set
  INSERT INTO #notehistory_set
    SELECT
      notehis_id,
      noteid,
      Action,
      Executer,
      TimeExecuted,
      event_type,
      name,
      Subject,
      Body,
      note_type,
      sevt_corr_val,
      twh_sevt_id,
      comp_id,
      twh_tevt_id,
      evt_corr_val
    FROM twh_defer_notehist
    WHERE comp_id <> 0
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err <> 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory: error %d inserting into #notehistory_set', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory inserted %d rows into temporary result set table', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- increment the age of the deferred notehistory so we know which were added
  -- this time in case TDW does not process this result set and we need
  -- to start over with the same beginning extraction id (last_xid)
  UPDATE twh_defer_notehist SET age = age + 1

  SELECT * FROM #notehistory_set

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getnotehistory returning', 0, 1, @now) WITH NOWAIT
  END

GO